<?php
/**
 * This file is part of the JustKnit package.
 *
 * @copyright JustKnit (Iris Hammer) <iris@justknit.de>
 * @license All rights reserved (@https://dejure.org/gesetze/UrhG/69a.html)
 * Created on 03-20-2023
 *
 * @package justknit
 */

namespace JustKnit\Controllers\Register;

use MODX\Revolution\modX;
use MODX\Revolution\modUser;
use MODX\Revolution\modUserProfile;
use JustKnit\Register\Register;
use JustKnit\Model\JustKnitClientMeta;
use JustKnit\Model\JustKnitProduct;
use JustKnit\Model\JustKnitOrder;

use MODX\Revolution\Mail\modPHPMailer;
use MODX\Revolution\Mail\modMail;

/**
 * Base controller class for register handling.
 *
 * @package justknit
 * @subpackage controllers
 */
abstract class Base {
    /** @var modX $modx */
    public $modx = null;

    /** @var modUser $user */
    public $user = null;

    /** @var modUserProfile $profile */
    public $profile = null;

    /** @var JustKnitClientMeta $clientmeta */
    public $clientmeta = null;

    /** @var Register $register */
    public $register = null;

    /** @var Dictionary $dictionary */
    public $dictionary = null;

    /** @var array $config An array of configuration properties */
    public $config = [];

    /** @var array $scriptProperties An array of script properties */
    protected $scriptProperties = [];

    /** @var JustKnitClientMeta.buyer_id $buyer_id */
    public $buyer_id = '';

    /** @var string $email */
    public $email = '';
    /** @var string $username */
    protected $username = '';
    /** @var string $password */
    protected $password = '';
    /** @var Register | Login | UpdateProfile | Calculation |  Project $controller */
    public $controller = null;

    /** @var boolean $inTestMode */
    public $inTestMode = false;    
    /**
     * The constructor for the Base controller class.
     *
     * @param Register $register A reference to the Register instance
     * @param array $config
     */
    public function __construct(Register &$register, array $config = [])
    {
        $this->modx = &$register->modx;
        $this->register = &$register;
        $this->config = array_merge($this->config, $config);
        $this->modx->lexicon->load('justknit:frontend');
        $this->modx->lexicon->load('justknit:register');
    }
    /**
     * Run a process.
     *
     * @param array $scriptProperties An array of properties
     * @access public
     */
    public function run(array $scriptProperties)
    {
        $this->setProperties($scriptProperties);
        $this->initialize();
        return $this->process();
    }
    /**
     * Abstracts initialize method.
     */
    abstract public function initialize();

    /**
     * Abstracts process method.
     */
    abstract public function process();
    /**
     * Check if an user with email address already exists.
     * MODX allow_multiple_emails setting is ignored -> we never let subscribe an email address more then once!
     *
     * @access public
     * @param string $email
     * @return mixed ID of MODX user || false
     */
    public function getUserByEmail(string $email)
    {
        $exists = false;
        $userProfile = $this->modx->getObject(modUserProfile::class, ['email' => $email]);
        if (is_object($userProfile)) {
            $exists = $userProfile->get('internalKey');
        }
        return $exists;
    }
    /**
     * Gets a user object by the MODX user ID.
     *
     * @access public
     * @params integer $userID
     * @return modUser object or null
     */
    public function getUserById(int $userID)
    {
        $this->user = $this->modx->getObject(modUser::class, [
            'id' => $userID,
        ]);
        if (!is_object($this->user)) {
            $this->user = null;
            $this->modx->log(modX::LOG_LEVEL_INFO, '[JustKnit] Could not load user with id: ' . $userID);
        }
        return $this->user;
    }

    /**
     * Gets a user object by the MODX user ID.
     *
     * @access public
     * @params integer $userID
     * @return modUser object or null
     */
    public function getClientByBuyerId(string $buyerID)
    {
        $this->clientmeta = $this->modx->getObject(JustKnitClientMeta::class, [
            'buyer_id' => $buyerID,
        ]);
        if (!is_object($this->clientmeta)) {
            $this->clientmeta = null;
            $this->modx->log(modX::LOG_LEVEL_INFO, '[JustKnit] Could not load client with buyer_id: ' . $buyerID);
        }
        return $this->clientmeta;
    }

    /**
     * Get the specified or active user
     * @return modUser
     */
    public function getUser() {
        $user = $this->getProperty('user',false);
        if (!empty($user)) {
            $c = intval($user) == 0 ? array('username' => $user) : intval($user);
            $this->user = $this->modx->getObject('modUser',$c);
        } else {
            $this->user =& $this->modx->user;
        }
        if (empty($this->user)) {
            $this->user = null;
            $this->modx->log(modX::LOG_LEVEL_ERROR,'[JustKnit] Could not find user: '.$user);
        }

        return $this->user;
    }
    /**
     * Get the Profile of the active user object.
     *
     * @access public
     * @return modUserProfile object or null
     */
    public function getProfile()
    {
        if (!is_object($this->user)) {
            $this->modx->log(modX::LOG_LEVEL_ERROR, '[JustKnit] User object to load profile doesn\'t exist.');
            return false;
        }
        $this->profile = $this->user->getOne('Profile');
        if (!is_object($this->profile)) {
            $this->profile = null;
            $this->modx->log(
                modX::LOG_LEVEL_INFO,
                '[JustKnit] Could not load profile for user: ' . $this->user->get('username')
            );
        }
        return $this->profile;
    }

    /**
     * Get ClientMeta of the active user object.
     *
     * @access public
     * @return JustKnitClientMeta object or null
     */
    public function getClientMeta()
    {
        if (!is_object($this->user)) {
            $this->modx->log(modX::LOG_LEVEL_ERROR, '[JustKnit] User object to load ClientMeta doesn\'t exist.');
            return false;
        }
        $this->clientmeta = $this->modx->getObject(JustKnitClientMeta::class, [
            'client_id' => $this->user->get('id'),
        ]);
        if (!is_object($this->clientmeta)) {
            $this->clientmeta = null;
            $this->modx->log(
                modX::LOG_LEVEL_INFO,
                '[JustKnit] Could not load JustKnitClientMeta for user with id: ' . $this->user->get('id')
            );
        }
        return $this->clientmeta;
    }
    /**
     * Gets a ClientMeta object by the MODX user ID.
     *
     * @access public
     * @params integer $userID
     * @return JustKnitClientMeta object or null
     */
    public function getClientMetaByUserId(int $userID)
    {
        $this->clientmeta = $this->modx->getObject(JustKnitClientMeta::class, [
            'client_id' => $userID,
        ]);
        if (!is_object($this->clientmeta)) {
            $this->clientmeta = null;
            $this->modx->log(
                modX::LOG_LEVEL_INFO,
                '[JustKnit] Could not load JustKnitClientMeta for user with id: ' . $userID
            );
        }
        return $this->clientmeta;
    }
    /**
     * Get product_id param from $_POST
     *
     * @access public
     * @return mixed string $product_id or false
     */
    public function getProductId()
    {
        if (isset($_POST['product_id'])) {
            $this->product_id = $_POST['product_id'];
            return $this->product_id;
        }
        if (isset($_GET['product_id'])) {
            $this->product_id = $_GET['product_id'];
            return $this->product_id;
        }        
        return false;
    }
    /**
     * Get Product object by the DS24 productId
     *
     * @access public
     * @params string $productId
     * @return JustKnitProduct object or null
     */
    public function getProduct()
    {
        $this->product = $this->modx->getObject(JustKnitProduct::class, [
            'product_id' => $this->product_id,
        ]);
        if (!is_object($this->product)) {
            $this->product = null;
            $this->modx->log(
                modX::LOG_LEVEL_INFO,
                '[JustKnit] Could not load JustKnitProduct for product with product_id: ' . $this->product_id
            );
        }
        return $this->product;
    }

    /**
     * Get order object by the DS24 order_id
     *
     * @access public
     * @params string $orderId
     * @return JustKnitOrder object or null
     */
    public function getOrder()
    {
        $this->order = $this->modx->getObject(JustKnitOrder::class, [
            'order_id' => $this->order_id,
        ]);
        if (!is_object($this->order)) {
            $this->order = null;
            $this->modx->log(
                modX::LOG_LEVEL_INFO,
                '[JustKnit] Could not load JustKnitOrder for product with order_id: ' . $this->order_id
            );
        }
        return $this->order;
    }    /**
     * Validate the posted sha_sign
     *
     * @access public
     * @return boolean
     */
    public function validateSha()
    {
        if ($this->hasPost()) {
            $param = $_POST;
        } elseif ($this->hasGet()) {
            $param = $_GET;
        } else {
            $param = [];
        }

        //$post = [];
        //$post['hasPost'] = false;
        $tyKey = $this->getThankYouPageKey();
        if (!empty($param)) {
            $ds24Resp = $this->modx->sanitize($param);
            //Only for debug
            unset($ds24Resp['login-updprof-btn']);
            $shaSign = $ds24Resp['sha_sign'];
            unset($ds24Resp['sha_sign']);
            $this->secureHash = $this->encodePost($ds24Resp, $tyKey);
            $this->modx->log( modX::LOG_LEVEL_ERROR,'[JustKnit] $properties '.$this->secureHash. print_r($param, 1));
            $this->shaIsValid = $this->shaIsValid($shaSign, $this->secureHash);
            $param['hasPost'] = true;
            $param = array_merge($ds24Resp, $param);
        }
        $this->setProperties($param);
        return $this->dictionary->gather($param);
    }

    /**
     * encode the posted array with ThankYouPageKey
     *
     * @access public
     * @return string
     */
    public function encodePost($post = [],string $tyKey ) {
        $keys = array_keys($post);
        sort($keys);
        $enc_string = '';

        foreach ($keys as $key) {
            $value = html_entity_decode( $post[ $key ] );
            $value = $post[ $key ];
            $is_empty = !isset($value) || $value === '' || $value === false;
            if ($is_empty)  continue;
            $enc_string .= "$key=$value$tyKey";
        }
        $enc_string = strtoupper(hash("sha512", $enc_string));
        $this->setProperty('enc_string', $enc_string);
        return $enc_string;
    }
    /**
     * Compare posted sha_sign with generated
     *
     * @access public
     * @return boolean
     */
    public function shaIsValid($shaSend, $shaEnc) {
        $isValid = false;
        if ($shaSend == $shaEnc) {
            $isValid = true;
        }
        return $isValid;
    }
    /**
     * Get the ThankYouPageKey from SystemSetting.
     *
     * @access public
     * @return string
     */
    public function getThankYouPageKey()
    {
        $this->tyPk = false;
        $this->tyPk = $this->modx->getOption('jk.ds24_typage_key');
        if (!$this->tyPk) {
            $this->modx->log(
                modX::LOG_LEVEL_INFO,
                '[JustKnit] Could not find ThankYouPageKey');
        }
        return $this->tyPk;
    }
    /**
     * Check if $_POST exists + load and set dictionary.
     *
     * @access public
     * @return boolean
     */
    public function hasPost()
    {
        $hasPost = false;
        if (!empty($_POST)) {
            $this->dictionary = $this->register->loadDictionary();
            if ($this->dictionary) {
                $hasPost = true;
            }
        }
        return $hasPost;
    }

    /**
     * Check if $_POST exists + load and set dictionary.
     *
     * @access public
     * @return boolean
     */
    public function hasGet()
    {
        $hasGet = false;
        if (!empty($_GET)) {
            $this->dictionary = $this->register->loadDictionary();
            if ($this->dictionary) {
                $hasGet = true;
            }
        }
        return $hasGet;
    }
    /**
     * trim prefix address from keys in scriptProperties.
     *
     * @access public
     * @return bool
     */
    public function trimAddress() {
        $props = $this->getProperties();
        if (!is_array($props)) {
            return false;
        }
        foreach ($props as $k => $v) {
            $k = explode("_", $k);
            if ($k[0] == 'address') {
                unset($k[0]);
            }
            $k = implode("_", $k);
            $props[$k] = $v;
        }
        unset($props['id']);
        $this->setProperties($props);
        return true;
    }
    /**
     * set properties related to duration of product client purchased.
     *
     * @access public
     * @return bool
     */
    public function setDurationValues() {
        $props = $this->getProperties();
        $props = array_merge([
            'blockedafter'    =>  strtotime(date("Y-m-d") . '+' . $props['duration']),
            'subscriptionstart' => time(),
            'subscriptionend' => strtotime(date("Y-m-d") . '+' . $props['duration']),
            'client_deletedafter' => strtotime(date("Y-m-d") . '+' . $props['duration'] . '+1month'),
            'active_product'  => $props['productID'],
            'DS24_product_id' => $props['product_id'],
        ],$props);
        $this->setProperties($props);
        return true;
    }
    /**
     * map posted properties to need of the table columns.
     *
     * @access public
     * @return bool
     */
    public function mapNewUserFields() {
        $props = $this->getProperties();
        if (!is_array($props)) {
            return false;
        }
        // map salutation
        if ($props['salutation']) {
            switch ($props['salutation']) {
                case 'F':
                    $props['gender'] = 2;
                    break;
                case 'M':
                    $props['gender'] = 1;
                    break;
            }
        } else {
                $props['gender'] = '';
        }
        $props = array_merge([
            'username'        => $props['email'],
            'phone'           => $props['phone_no'],
            'mobilephone'     => $props['mobile_no'],
            'address'         => $props['street'],
            'zip'             => $props['zipcode'],
            'nick_name'       => $props['first_name'],
            'fullname'        => $props['first_name'] . ' ' . $props['last_name'],
            'createdon'       => time(),
        ],$props);
        $this->setProperties($props);
        return true;
    }
    /**
     * Get message of error response.
     *
     * @access public
     * @return json
     */
    public function getJsonError(string $message)
    {
        $this->response['status'] = 'error';
        $this->response['message'] = $message;
        return json_encode($this->response);
    }
    /**
     * Get message of success response.
     *
     * @access public
     * @return json
     */
    public function getJsonSuccess()
    {
        $this->response['status'] = 'success';
        return json_encode($this->response);
    }
    /**
     * Set the default script properties.
     *
     * @param array $defaults
     * @return void
     */
    function setDefaultProperties(array $defaults = [])
    {
        $this->scriptProperties = array_merge($defaults, $this->scriptProperties);
    }
    /**
     * Set a script property.
     *
     * @access public
     * @param string $key
     * @param mixed $value
     * @return void
     */
    public function setProperty(string $key, $value)
    {
        $this->scriptProperties[$key] = $value;
    }
    /**
     * Set an array of script properties.
     *
     * @access public
     * @param array $array
     * @return void
     */
    public function setProperties(array $array)
    {
        foreach ($array as $k => $v) {
            $this->setProperty($k, $v);
        }
    }
    /**
     * Get a script property.
     *
     * @access public
     * @param string $key
     * @param mixed $default
     * @param string $method
     * @return mixed
     */
    public function getProperty(string $key, $default = null, string $method = '!empty')
    {
        $v = $default;

        switch ($method) {
            case 'empty':
            case '!empty':
                if (!empty($this->scriptProperties[$key])) {
                    $v = $this->scriptProperties[$key];
                }
                break;

            case 'isset':
            default:
                if (isset($this->scriptProperties[$key])) {
                    $v = $this->scriptProperties[$key];
                }
                break;
        }
        return $v;
    }

    /**
     * Return an array of REQUEST script properties.
     *
     * @access public
     * @return array
     */
    public function getProperties()
    {
        return $this->scriptProperties;
    }

    /**
     * Set a placeholder.
     *
     * @access public
     * @param mixed $k
     * @param mixed $v
     * @return void
     */
    public function setPlaceholder($k, $v)
    {
        $this->placeholders[$k] = $v;
    }

    /**
     * Get a placeholder.
     *
     * @access public
     * @param mixed $k
     * @param mixed $default (default: null)
     * @return void
     */
    public function getPlaceholder($k, $default = null)
    {
        return isset($this->placeholders[$k]) ? $this->placeholders[$k] : $default;
    }

    /**
     * Set an array of placeholders.
     *
     * @access public
     * @param array $array
     * @return void
     */
    public function setPlaceholders(array $array)
    {
        foreach ($array as $k => $v) {
            $this->setPlaceholder($k, $v);
        }
    }

    /**
     * Get an array of placeholders.
     *
     * @access public
     * @return void
     */
    public function getPlaceholders()
    {
        return $this->placeholders;
    }
    /**
     * Run a desired processor.
     *
     * @access public
     * @param string $processor Name of the processor to run
     * @return mixed|string
     */
    public function runProcessor(string $processor)
    {
        $output = '';
        $processor = $this->loadProcessor($processor);
        if (empty($processor)) {
            return $output;
        }
        // Return the output of the processor
        return $processor->process();

    }

    /**
     * Loads a processor.
     *
     * @access public
     * @param string $processor Name of the processor to load
     * @return bool|Subscription processor
     */
    public function loadProcessor(string $processor)
    {
        $processorFile = $this->config['processorsPath'] . $processor . '.php';
$this->modx->log( modX::LOG_LEVEL_ERROR,'[JustKnit] $processorFile '.$processorFile);

        if (!file_exists($processorFile)) {
            $this->modx->log(modX::LOG_LEVEL_ERROR, '[JustKnit] Could not find processor file: ' . $processorFile);
            return false;
        }
        try {
            $className = 'JustKnit\\Processors\\Register\\' . $processor;
$this->modx->log( modX::LOG_LEVEL_ERROR,'[JustKnit] $className, $processor '.$className.', '.$processor);
            if (!class_exists($className)) {
                $className = include_once $processorFile;
            }
            if ($processor == 'ClientLogin') {
                $props = $this->register->dictionary->toArray();

$this->modx->log( modX::LOG_LEVEL_ERROR,'[JustKnit] $props '.print_r($props,1));
                $processor = new $className($this->modx, $props);
            } else {

            $processor = new $className($this->register, $this);
            }

        } catch (\Exception $e) {
            $processor = false;
            $this->modx->log(modX::LOG_LEVEL_ERROR, '[JustKnit] ' . $e->getMessage());
        }
        return $processor;
    }

    /**
     * Manipulate/add/remove fields from array.
     *
     * @access protected
     * @param array $properties
     * @return array $properties The cleaned array
     */
    protected function cleanupKeys(array $properties = [])
    {
        unset(
            // users table
            $properties['id'],          // multiple occurrence; not needed
            $properties['password'],    // security!
            $properties['cachepwd'],    // security!
            $properties['hash_class'],  // security!
            $properties['salt'],        // security!
            // user_attributes table
            $properties['internalKey'], // not needed
            $properties['sessionid'],   // security!
            $properties['extended'],     // not needed as its already flattened
            $properties['comment']
        );
        return $properties;
    }

    /**
     * escapePlaceholders from array.
     * core/components/login/model/login/logincontroller.class.php:142
     */
    public function escapePlaceholders($val) {
        if (is_array($val)) {
            $val = array_map(array($this, 'escapePlaceholders'), $val);
        } else {
            $val = htmlspecialchars($val, ENT_QUOTES, $this->modx->getOption('modx_charset', null, 'UTF-8'));
            $val = str_replace(array('[', ']'), array('&#91;', '&#93;'), $val);
        }
        return $val;
    }

    /**
     * Gets the current Request URI.
     *
     * @access public
     * @return string The request URI, with Login-specific code stripped.
     */
    public function getRequestURI() {
        $requestUri = php_sapi_name() == 'cli' ? '' : $_SERVER['REQUEST_URI'];
        return str_replace(array('?service=logout','&service=logout'),'',$requestUri);
    }

    /**
     * Gets a user by a specific field in the table.
     *
     * @access public
     * @param string $field The column to grab by.
     * @param string $value The value to search by.
     * @param string $alias Optional; allows searching by a separate class
     * alias. Defaults to modUser.
     * @return modUser/null Returns a modUser object if successfull; null if
     * not.
     */
    public function getUserByField($field,$value,$alias = modUser::class) {
        $c = $this->modx->newQuery(modUser::class);
        $c->innerJoin(modUserProfile::class,'Profile');
        $c->select($this->modx->getSelectColumns(modUser::class,'modUser'));
        $c->select(array(
            'email' => 'Profile.email',
        ));
        $c->where(array(
            $alias.'.'.$field => $value,
        ));
        return $this->modx->getObject(modUser::class,$c);
    }

    /**
     * Encodes a string for URL safe transmission
     *
     * @access public
     * @param string $str
     * @return string
     */
    public function base64url_encode($str) { 
        return rtrim(strtr(base64_encode($str), '+/', '-_'), '='); 
    }    
    
    /**
     * Decodes an URL safe encoded string
     *
     * @access public
     * @param string $str
     * @return string
     */
    public function base64url_decode($str) { 
        return base64_decode(str_pad(strtr($str, '-_', '+/'), strlen($str) % 4, '=', STR_PAD_RIGHT)); 
    }
    /**
     * Sends an email based on the specified information and templates.
     *
     * @access public
     * @param string $email The email to send to.
     * @param string $name The name of the user to send to.
     * @param string $subject The subject of the email.
     * @param array $properties A collection of properties.
     * @return array
     */
    public function sendEmail($email,$name,$subject,$properties = array()) {
        if (empty($properties['tpl'])) $properties['tpl'] = 'jkForgotPassEmail';
        if (empty($properties['tplType'])) $properties['tplType'] = 'modChunk';

        $msg = $this->modx->getChunk($properties['tpl'],$properties,$properties['tplType']);
        $msgAlt = '';
        if (!empty($properties['tplAlt'])) {
            $msgAlt = $this->modx->getChunk($properties['tplAlt'],$properties,$properties['tplType']);
        }
        // New for MODX3
        if (!$this->modx->services->has('mail')) {
            $this->modx->services->add('mail', new modPHPMailer($this->modx));
        }
        $mail = $this->modx->services->get('mail');
$this->modx->log(modX::LOG_LEVEL_ERROR, 'Class of \$mail: ' . get_class($mail));
$this->modx->log(modX::LOG_LEVEL_ERROR, 'Class of \$mail->mailer: ' . get_class($mail->mailer));
//        $this->modx->getService('mail', 'mail.modPHPMailer');
        $mail->set(modMail::MAIL_BODY, $msg);
        $mail->set(modMail::MAIL_FROM, $this->modx->getOption('emailFrom', $properties, $this->modx->getOption('emailsender'), true));
        $mail->set(modMail::MAIL_FROM_NAME, $this->modx->getOption('emailFromName', $properties, $this->modx->getOption('site_name'), true));
        $mail->set(modMail::MAIL_SENDER, $this->modx->getOption('emailSender', $properties, $this->modx->getOption('emailsender'), true));
        $mail->set(modMail::MAIL_SUBJECT, $subject);
        if (!empty($msgAlt)) {
            $mail->set(modMail::MAIL_BODY_TEXT, $msgAlt);
        }
        $mail->address('to', $email, $name);
        $replyTo = $this->modx->getOption('emailReplyTo', $properties);
        if (!empty($replyTo)) {
            $mail->address('reply-to', $replyTo);
        }
        $mail->setHTML(true);
        if ($this->inTestMode) return true;
        $sent = $mail->send();
        if (!$sent) {
            $this->modx->log(modX::LOG_LEVEL_ERROR, '[JustKnit] '.$this->modx->lexicon('justknit.email_not_sent').' '.print_r($this->modx->mail->mailer->ErrorInfo, true));
        }
        $mail->reset();

        return $sent;
    }    
}
